@@ -17,9 +17,9 @@
 
 #include "serf.h"
 #include "serf_bucket_util.h"
+#include "serf_private.h"
 
-
-SERF_DECLARE(serf_bucket_t *) serf_bucket_create(
+serf_bucket_t *serf_bucket_create(
     const serf_bucket_type_t *type,
     serf_bucket_alloc_t *allocator,
     void *data)
@@ -33,7 +33,8 @@ SERF_DECLARE(serf_bucket_t *) serf_bucket_create(
     return bkt;
 }
 
-SERF_DECLARE(apr_status_t) serf_default_read_iovec(
+
+apr_status_t serf_default_read_iovec(
     serf_bucket_t *bucket,
     apr_size_t requested,
     int vecs_size,
@@ -69,7 +70,8 @@ SERF_DECLARE(apr_status_t) serf_default_read_iovec(
     return status;
 }
 
-SERF_DECLARE(apr_status_t) serf_default_read_for_sendfile(
+
+apr_status_t serf_default_read_for_sendfile(
     serf_bucket_t *bucket,
     apr_size_t requested,
     apr_hdtr_t *hdtr,
@@ -94,14 +96,16 @@ SERF_DECLARE(apr_status_t) serf_default_read_for_sendfile(
     return status;
 }
 
-SERF_DECLARE(serf_bucket_t *) serf_default_read_bucket(
+
+serf_bucket_t *serf_default_read_bucket(
     serf_bucket_t *bucket,
     const serf_bucket_type_t *type)
 {
     return NULL;
 }
 
-SERF_DECLARE(void) serf_default_destroy(serf_bucket_t *bucket)
+
+void serf_default_destroy(serf_bucket_t *bucket)
 {
 #ifdef SERF_DEBUG_BUCKET_USE
     serf_debug__bucket_destroy(bucket);
@@ -110,34 +114,20 @@ SERF_DECLARE(void) serf_default_destroy(serf_bucket_t *bucket)
     serf_bucket_mem_free(bucket->allocator, bucket);
 }
 
-SERF_DECLARE(void) serf_default_destroy_and_data(serf_bucket_t *bucket)
+
+void serf_default_destroy_and_data(serf_bucket_t *bucket)
 {
     serf_bucket_mem_free(bucket->allocator, bucket->data);
     serf_default_destroy(bucket);
 }
 
-SERF_DECLARE(apr_status_t) serf_default_snapshot(serf_bucket_t *bucket)
-{
-    /* Not implemented. */
-    return APR_ENOTIMPL;
-}
-
-SERF_DECLARE(apr_status_t) serf_default_restore_snapshot(serf_bucket_t *bucket)
-{
-    /* Not implemented. */
-    return APR_ENOTIMPL;
-}
-
-SERF_DECLARE(int) serf_default_is_snapshot_set(serf_bucket_t *bucket)
-{
-    return 0;  
-}
 
 /* ==================================================================== */
 
 
-SERF_DECLARE(char *) serf_bstrmemdup(serf_bucket_alloc_t *allocator,
-                                     const char *str, apr_size_t size)
+char *serf_bstrmemdup(serf_bucket_alloc_t *allocator,
+                      const char *str,
+                      apr_size_t size)
 {
     char *newstr = serf_bucket_mem_alloc(allocator, size + 1);
     memcpy(newstr, str, size);
@@ -145,17 +135,19 @@ SERF_DECLARE(char *) serf_bstrmemdup(serf_bucket_alloc_t *allocator,
     return newstr;
 }
 
-SERF_DECLARE(void *) serf_bmemdup(serf_bucket_alloc_t *allocator,
-                                  const void *mem,
-                                  apr_size_t size)
+
+void *serf_bmemdup(serf_bucket_alloc_t *allocator,
+                   const void *mem,
+                   apr_size_t size)
 {
     void *newmem = serf_bucket_mem_alloc(allocator, size);
     memcpy(newmem, mem, size);
     return newmem;
 }
 
-SERF_DECLARE(char *) serf_bstrdup(serf_bucket_alloc_t *allocator,
-                                  const char *str)
+
+char *serf_bstrdup(serf_bucket_alloc_t *allocator,
+                   const char *str)
 {
     apr_size_t size = strlen(str) + 1;
     char *newstr = serf_bucket_mem_alloc(allocator, size);
@@ -163,6 +155,32 @@ SERF_DECLARE(char *) serf_bstrdup(serf_bucket_alloc_t *allocator,
     return newstr;
 }
 
+char *serf_bstrcatv(serf_bucket_alloc_t *allocator, struct iovec *vec,
+                    int vecs, apr_size_t *bytes_written)
+{
+    int i;
+    apr_size_t new_len = 0;
+    char *c, *newstr;
+
+    for (i = 0; i < vecs; i++) {
+        new_len += vec[i].iov_len;
+    }
+
+    /* It's up to the caller to free this memory later. */
+    newstr = serf_bucket_mem_alloc(allocator, new_len);
+
+    c = newstr;
+    for (i = 0; i < vecs; i++) {
+        memcpy(c, vec[i].iov_base, vec[i].iov_len);
+        c += vec[i].iov_len;
+    }
+
+    if (bytes_written) {
+        *bytes_written = c - newstr;
+    }
+
+    return newstr;
+}
 
 /* ==================================================================== */
 
@@ -203,8 +221,12 @@ static void find_crlf(const char **data, apr_size_t *len, int *found)
     *found = SERF_NEWLINE_NONE;
 }
 
-SERF_DECLARE(void) serf_util_readline(const char **data, apr_size_t *len,
-                                      int acceptable, int *found)
+
+void serf_util_readline(
+    const char **data,
+    apr_size_t *len,
+    int acceptable,
+    int *found)
 {
     const char *start;
     const char *cr;
@@ -284,7 +306,7 @@ SERF_DECLARE(void) serf_util_readline(const char **data, apr_size_t *len,
 /* ==================================================================== */
 
 
-SERF_DECLARE(void) serf_databuf_init(serf_databuf_t *databuf)
+void serf_databuf_init(serf_databuf_t *databuf)
 {
     /* nothing is sitting in the buffer */
     databuf->remaining = 0;
@@ -326,10 +348,12 @@ static apr_status_t common_databuf_prep(serf_databuf_t *databuf,
     return APR_SUCCESS;
 }
 
-SERF_DECLARE(apr_status_t) serf_databuf_read(serf_databuf_t *databuf,
-                                             apr_size_t requested,
-                                             const char **data,
-                                             apr_size_t *len)
+
+apr_status_t serf_databuf_read(
+    serf_databuf_t *databuf,
+    apr_size_t requested,
+    const char **data,
+    apr_size_t *len)
 {
     apr_status_t status = common_databuf_prep(databuf, len);
     if (status)
@@ -355,10 +379,13 @@ SERF_DECLARE(apr_status_t) serf_databuf_read(serf_databuf_t *databuf,
     return databuf->remaining ? APR_SUCCESS : databuf->status;
 }
 
-SERF_DECLARE(apr_status_t) serf_databuf_readline(serf_databuf_t *databuf,
-                                                 int acceptable, int *found,
-                                                 const char **data,
-                                                 apr_size_t *len)
+
+apr_status_t serf_databuf_readline(
+    serf_databuf_t *databuf,
+    int acceptable,
+    int *found,
+    const char **data,
+    apr_size_t *len)
 {
     apr_status_t status = common_databuf_prep(databuf, len);
     if (status)
@@ -378,9 +405,11 @@ SERF_DECLARE(apr_status_t) serf_databuf_readline(serf_databuf_t *databuf,
     return databuf->remaining ? APR_SUCCESS : databuf->status;
 }
 
-SERF_DECLARE(apr_status_t) serf_databuf_peek(serf_databuf_t *databuf,
-                                             const char **data,
-                                             apr_size_t *len)
+
+apr_status_t serf_databuf_peek(
+    serf_databuf_t *databuf,
+    const char **data,
+    apr_size_t *len)
 {
     apr_status_t status = common_databuf_prep(databuf, len);
     if (status)
@@ -403,13 +432,14 @@ SERF_DECLARE(apr_status_t) serf_databuf_peek(serf_databuf_t *databuf,
 /* ==================================================================== */
 
 
-SERF_DECLARE(void) serf_linebuf_init(serf_linebuf_t *linebuf)
+void serf_linebuf_init(serf_linebuf_t *linebuf)
 {
     linebuf->state = SERF_LINEBUF_EMPTY;
     linebuf->used = 0;
 }
 
-SERF_DECLARE(apr_status_t) serf_linebuf_fetch(
+
+apr_status_t serf_linebuf_fetch(
     serf_linebuf_t *linebuf,
     serf_bucket_t *bucket,
     int acceptable)
@@ -460,11 +490,10 @@ SERF_DECLARE(apr_status_t) serf_linebuf_fetch(
 
                 /* Whatever was read, the line is now ready for use. */
                 linebuf->state = SERF_LINEBUF_READY;
+            } else {
+                /* no data available, try again later. */
+                return APR_EAGAIN;
             }
-            /* ### we need data. gotta check this char. bail if zero?! */
-            /* else len == 0 */
-
-            /* ### status */
         }
         else {
             int found;
@@ -532,3 +561,80 @@ SERF_DECLARE(apr_status_t) serf_linebuf_fetch(
     }
     /* NOTREACHED */
 }
+
+/* Logging functions.
+   Use with one of the [COMP]_VERBOSE defines so that the compiler knows to
+   optimize this code out when no logging is needed. */
+static void log_time()
+{
+    apr_time_exp_t tm;
+
+    apr_time_exp_lt(&tm, apr_time_now());
+    fprintf(stderr, "[%d-%02d-%02dT%02d:%02d:%02d.%06d%+03d] ",
+            1900 + tm.tm_year, 1 + tm.tm_mon, tm.tm_mday,
+            tm.tm_hour, tm.tm_min, tm.tm_sec, tm.tm_usec,
+            tm.tm_gmtoff/3600);
+}
+
+void serf__log(int verbose_flag, const char *filename, const char *fmt, ...)
+{
+    va_list argp;
+
+    if (verbose_flag) {
+        log_time();
+
+        if (filename)
+            fprintf(stderr, "%s: ", filename);
+
+        va_start(argp, fmt);
+        vfprintf(stderr, fmt, argp);
+        va_end(argp);
+    }
+}
+
+void serf__log_nopref(int verbose_flag, const char *fmt, ...)
+{
+    va_list argp;
+
+    if (verbose_flag) {
+        va_start(argp, fmt);
+        vfprintf(stderr, fmt, argp);
+        va_end(argp);
+    }
+}
+
+void serf__log_skt(int verbose_flag, const char *filename, apr_socket_t *skt,
+                   const char *fmt, ...)
+{
+    va_list argp;
+
+    if (verbose_flag) {
+        apr_sockaddr_t *sa;
+        log_time();
+
+        if (skt) {
+            /* Log local and remote ip address:port */
+            fprintf(stderr, "[l:");
+            if (apr_socket_addr_get(&sa, APR_LOCAL, skt) == APR_SUCCESS) {
+                char buf[32];
+                apr_sockaddr_ip_getbuf(buf, 32, sa);
+                fprintf(stderr, "%s:%d", buf, sa->port);
+            }
+            fprintf(stderr, " r:");
+            if (apr_socket_addr_get(&sa, APR_REMOTE, skt) == APR_SUCCESS) {
+                char buf[32];
+                apr_sockaddr_ip_getbuf(buf, 32, sa);
+                fprintf(stderr, "%s:%d", buf, sa->port);
+            }
+            fprintf(stderr, "] ");
+        }
+
+        if (filename)
+            fprintf(stderr, "%s: ", filename);
+
+        va_start(argp, fmt);
+        vfprintf(stderr, fmt, argp);
+        va_end(argp);
+    }
+}
+
